@@ -24,10 +24,8 @@ application ID can also be used to generate a URL or email address, or to make a
 run-time decision.
 
 The application ID is the same as the project ID. To learn more about the App
-Identity API, read the `App Identity Python Overview`_.
-
-.. _App Identity Python Overview:
-   https://cloud.google.com/appengine/docs/python/appidentity/
+Identity API, read the App Identity Python Overview:
+https://cloud.google.com/appengine/docs/python/appidentity/
 """
 
 from google.appengine.api.app_identity.app_identity import *
